/*
 * $Log: milter.c,v $
 * Revision 1.3  2004/09/14 11:14:24  mi
 * Use intptr_t instead of int, when casting the result of skem_cleanup() to
 * satisfy 64-bit platforms.
 *
 * Revision 1.2  2004/07/14 05:09:47  mi
 * Prepare for the 1.0 release
 *
 * Revision 1.1  2004/06/29 20:55:33  mi
 * Initial revision
 *
 */
#include <errno.h>
#include <fcntl.h>
#include <libmilter/mfapi.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pwd.h>
#include <regex.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>

#define SKEM_MAIN_PROGRAM
#include "skem.h"
#undef SKEM_MAIN_PROGRAM

#ifndef SKEM_DEFAULT_GRACE
#	define SKEM_DEFAULT_GRACE	1800	/* 30 minutes */
#endif

struct skem_options O = {
	sizeof(struct skem_options),
	SKEM_DEFAULT_GRACE,	/* how long to keep the temp-bans */
	SMFIS_CONTINUE,		/* what to return when seeing a ban */
	0,			/* uid to run as -- milter will not be root */
	".",			/* working directory -- "." */
	NULL,			/* socket path -- must be specified */
	NULL,			/* PID filename -- optional */
#ifdef SKEM_LOGWATCHER
	NULL,			/* Logwatcher thread id */
	"reject=5",		/* sendmail rejected something */
	"\\[([0-9]+\\.[0-9]+\\.[0-9]+\\.[0-9]+)\\]", /* IPv4 patten XXX */
	NULL,			/* spamword -- to issue permanent bans */
#endif
#ifdef SKEM_CLEANUP	
	NULL,			/* Cleanup thread id */
	3*7*24*60*60,		/* when to remove perm-bans -- 3 weeks */
	{ 0 },			/* how often to cleanup */
#endif
};

static const char *version = "$Id: milter.c,v 1.3 2004/09/14 11:14:24 mi Exp mi $"
    " compiled on " __DATE__ " at "__TIME__;

static sfsistat helo(SMFICTX *ctx, char *helohost);
static sfsistat conn(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr);
static sfsistat cleanup(SMFICTX *ctx);
static sfsistat eoh(SMFICTX *ctx);

static struct smfiDesc smfilter = {
	"skem",		/* filter name */
	SMFI_VERSION,	/* version code -- do not change */
	0,		/* flags */
	conn,		/* connection info filer */
	helo,		/* SMTP HELO command filter */
	NULL,		/* envelope sender filter */
	NULL,		/* envelope recipient filter */
	NULL,		/* header filter */
	NULL,		/* end of header */
	NULL,		/* body block filter */
	NULL,		/* end of message */
	cleanup,	/* message aborted */		
	cleanup,	/* connection cleanup */
};

int
skem_isanIP(const char *s)
{
	char	buf[8];

	return (
#ifdef NETINET
	    inet_pton(AF_INET, s, &buf) == 1
#	ifdef NETINET6
		||
#	endif
#endif
#ifdef NETINET6
	    inet_pton(AF_INET6, s, &buf) == 1
#endif
	    );
}

static int
check(SMFICTX *ctx, const char *client_addr)
{
	struct stat	 sb;
	char		 mtime[26];
	struct timeval	 tp;
	long		 delta;
	const char	*queueid;
	pthread_t	 myself;
	
	queueid = smfi_getsymval(ctx, "{i}");
	myself = pthread_self();

	if (queueid == NULL)
		queueid = "(SKEM_NOQUEU)";

	if (lstat(client_addr, &sb) != 0) {
		if (errno != ENOENT)
			syslog(LOG_ERR, "Checking %s: %m", client_addr);
		return SMFIS_CONTINUE; 
	}

	if (sb.st_mode & S_ISUID) {
		syslog(LOG_INFO, "%s is permanently whitelisted "
		    "(setuid)", client_addr);
		return SMFIS_CONTINUE;
	}

	if (sb.st_mode & S_ISGID) {

		ctime_r(&sb.st_mtime, mtime);
		if (difftime(sb.st_ctime, sb.st_mtime) < 1)
			syslog(LOG_INFO, "%s: %s is permanently banned"
			    " (setgid) since %.24s", queueid,
			    client_addr, mtime);
		else {
			char ctime[26];

			ctime_r(&sb.st_ctime, ctime);
			syslog(LOG_INFO, "%s: %s is permanently banned"
			    " (setgid) since %.24s"
			    "; last seen on %.24s", queueid,
			    client_addr, mtime, ctime);
		}

		/* Update the file's ctime */
		if (chmod(client_addr, sb.st_mode))
			syslog(LOG_ERR, "Updating ctime on %s: %m",
			    client_addr);
		return SMFIS_REJECT;
	}

	gettimeofday(&tp, NULL);
	delta = tp.tv_sec - sb.st_mtime;

	if (delta < O.grace) {
		syslog(LOG_INFO, "%s: %p %s was banned %ld seconds ago. "
		    "Rejecting", queueid, myself, client_addr, delta);
		return SMFIS_TEMPFAIL;
	}

	ctime_r(&sb.st_mtime, mtime);
	syslog(LOG_INFO, "%s: %s was banned %ld seconds ago "
	    "on %.24s. Removing the expired ban",
	    queueid, client_addr, delta, mtime);
	if (unlink(client_addr))
		syslog(LOG_ERR, "can not unban %s: %m",
		    client_addr);

	return SMFIS_CONTINUE;
}

static sfsistat
conn(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr)
{
	char		*client_addr;
	int		 len;
	const void	*addr;

	/* syslog(LOG_DEBUG, "in %s", __func__); */

	if (hostaddr == NULL)
		return SMFIS_CONTINUE;

	switch(hostaddr->sa_family) {
#ifdef NETINET
	case AF_INET:
		addr = &((struct sockaddr_in *)hostaddr)->sin_addr;
		len = INET_ADDRSTRLEN;
		break;
#endif
#ifdef NETINET6
	case AF_INET6:
		addr = &((struct sockaddr_in6 *)hostaddr)->sin6_addr;
		len = INET6_ADDRSTRLEN;
		break;
#endif
	default:
		syslog(LOG_ERR, "Unrecognized address family %d for %s",
		    hostaddr->sa_family, hostname);
		return SMFIS_CONTINUE;
	}

	client_addr = malloc(len);
	if (client_addr == NULL) {
		syslog(LOG_CRIT, "Out of memory");
		return SMFIS_TEMPFAIL;
	}

	if (inet_ntop(hostaddr->sa_family, addr, client_addr, len) == NULL) {
		syslog(LOG_ERR, "extracting IP address from %p for %s: %m",
		    hostaddr, hostname);
		return SMFIS_CONTINUE;
	} else
		syslog(LOG_DEBUG, "connection from %s (%s)", hostname,
		    client_addr);

	switch (check(ctx, client_addr)) {
	case SMFIS_REJECT:
		free(client_addr);
		return SMFIS_REJECT;
	case SMFIS_CONTINUE:
		if (smfilter.xxfi_eoh == eoh)
			smfi_setpriv(ctx, client_addr);
		else
			free(client_addr);
		return SMFIS_CONTINUE;
	case SMFIS_TEMPFAIL:
		if (O.tempfail == SMFIS_CONTINUE) {
			smfi_setpriv(ctx, client_addr);
			return SMFIS_CONTINUE; /* helo() will reject */
		}
		
		/*
		 * The following text is ignored by the current
		 * libmilter, but the string "421" is vital
		 */
		smfi_setreply(ctx, "421", "4.1.8", "you are "
		    "temporarily blacklisted from this system");
		free(client_addr);
		return SMFIS_TEMPFAIL;
	default:
		syslog(LOG_CRIT, "Unexpected result from check() on %s. "
		    "Exiting in shame", client_addr);
		exit(EX_SOFTWARE);
	}
}

/*
 * This function is only used if temporary rejections directly from
 * conn() are not possible. In this case, conn() is still making all
 * of the decisions, but instead of returning SMFIS_TEMPFAIL (as we
 * would otherwise prefer), returns SMFIS_CONTINUE after setting the
 * private data to non-NULL. helo() checks for that setting, and returns
 * TEMPFAIL if asked. Neither this function, nor the cleanup() below
 * it are ever invoked, if we are started with the `-4' option, which
 * implies, the sendmail was compiled with -D_FFR_MILTER_421 and will
 * understand SMFIS_TEMPFAIL from conn() as we want it to.
 */

static sfsistat
helo(SMFICTX *ctx, char *helohost)
{

	if (smfi_getpriv(ctx)) {
		smfi_setreply(ctx, "421", "4.1.8", "you are "
		    "temporarily blacklisted from this system");
		return SMFIS_TEMPFAIL;
	}
	return SMFIS_CONTINUE;
}

static sfsistat
eoh(SMFICTX *ctx)
{
	const char 	*client_addr;
	sfsistat	 result;

	client_addr = smfi_getpriv(ctx);
	if (client_addr == NULL) {
		syslog(LOG_ERR, "eoh called, but no relay address recorded");
		return SMFIS_CONTINUE;
	}
	result = check(ctx, client_addr);
	if (result != SMFIS_CONTINUE) {
		syslog(LOG_INFO, "conn/helo let %s slip, but it was "
		    "determined suspicious by the end of headers",
		    client_addr);
	}
	return result;
}

static sfsistat
cleanup(SMFICTX *ctx)
{
	void *p;

	p = smfi_getpriv(ctx);
	if (p) {
		free(p);
		smfi_setpriv(ctx, NULL);
	}

	return SMFIS_CONTINUE;
}

int
skem_touch(const char *name, int mode)
{
	int result;

	result = open(name, O_WRONLY | O_CREAT, mode);
	if (result != -1)
		result = close(result);

	return result;
}

static int
list(int argc, char * const argv[], int mode)
{
	struct stat	sb;
	int	e = 0;

	for (; argc--; argv++) {
		if (!skem_isanIP(*argv)) {
			syslog(LOG_ERR, "%slisting: %s is not an IP-address",
			    mode & S_ISGID ? "Black" : "White", *argv);
			continue;
		}
		if (lstat(*argv, &sb) == 0) {
			syslog(LOG_WARNING, "%slisting: %s already exists, "
			    "skipping",
			    mode & S_ISGID ? "Black" : "White",
			    *argv);
			continue;
		}
		if (errno != ENOENT) {
			syslog(LOG_ERR, "%slisting: can't lstat %s: "
			    "%m", mode & S_ISGID ? "Black" : "White", *argv);
			e = errno;
		} else if (skem_touch(*argv, mode) == -1) {
			syslog(LOG_ERR, "%slisting: can't add %s: "
			    "%m", mode & S_ISGID ? "Black" : "White", *argv);
			e = errno;
		} else
			syslog(LOG_DEBUG, "%slister: added %s",
			    mode & S_ISGID ? "Black" : "White", *argv);
	}

	if (e)
		return e == EPERM ? EX_NOPERM : EX_DATAERR;
	else
		return 0;
}

static void
usage(const char *prog, int exit_code)
{
	fprintf(stderr, "Usage:\n\t%s -v (print version and exit) or\n"
	    "\t[-C dbdirectory]\n"
	    "\t[-u uid (required if started by root)]\n"
	    "\t[-w ip1 [ip2 ... ipN] -- whitelist the ips and exit] or\n"
	    "\t[-b ip1 [ip2 ... ipN] -- blacklist the ips and exit] or\n"
	    "\t[-d (allow LOG_DEBUG rated message)\n"
	    "\t[-g graceperiod (in seconds)]\n"
	    "\t[-p pidfile (optional)]\n"
#ifdef SKEM_CLEANUP
	    "\t[-B seconds (when to clean-up untriggered permanent bans "
		"(%u))]\n"
	    "\t[-l dbcleanupperiod (in seconds -- optional)]\n"
	    "\t[-L (just do cleanups -- once if -l is not given)]\n"
#else
	    "\t[-B|-L|-l -- cleanup not compiled int]\n"
#endif
	    "\t[-2 (to do the second check at the end of headers)]\n"
	    "\t[-4 (if you are sure, sendmail was compiled with "
		"-D_FFR_MILTER_421)]\n"
#ifdef SKEM_LOGWATCHER
	    "\t[-S spamword (a string, indicating a definite spam-source)]\n"
	    "\t[-s pattern (if standard input contains log messagesl\n"
	    "\t\tuse ``-'' for the builtin ``%s'')]\n"
	    "\t[-r reject-string (if the ``%s'' is not good for you;\n"
	    "\t\tuse ``-'' to pass all stdin through regexp)]\n"
#else
	    "\t[-r|-S|-s -- logwatcher not compiled in\n"
#endif
	    "\tsocketpath\n\n"
	    "Where socketpath is passed to smfi_setconn() and is one of the\n"
	    "\t{unix|local}:/path/to/file -- a named pipe\n"
#if NETINET
	    "\tinet:port@{hostname|ip-address} -- an IPV4 socket\n"
#endif
#if NETINET6
	    "\tinet6:port@{hostname|ip-address} -- an IPV6 socket\n"
#endif
	    "\nDefault grace period is %u, currently set to %u.\n"
	    "\n"
	    "As the switches are processed in the order given, you can\n"
	    "change some of the behavior by placing the ``-u'' option\n"
	    "appropriately. For example, placing it after the ``-p'',\n"
	    "will cause the PID to be created by root (and then chown-ed)\n"
	    "Also, the ``-p'' and ``-P'' switches may interpreted differently\n"
	    "depending on their position relative to the ``-C''.\n",
	    prog,
#ifdef SKEM_CLEANUP
	    O.permcleanup,
#endif
#ifdef SKEM_LOGWATCHER
	    O.ippattern, O.reject,
#endif
	    SKEM_DEFAULT_GRACE, O.grace);
	exit(exit_code);
}

static void
exiting(void)
{

	if (O.pidfname)
		if (unlink(O.pidfname))
			syslog(LOG_WARNING, "Could not unlink my pid file "
			    "%s on exit: %m", O.pidfname);
#if 0
	if (O.socket)
		if (unlink(O.socket))
			syslog(LOG_WARNING, "Could not unlink the socket "
			    "%s on exit: %m", O.socket);
#endif
	O.pidfname = O.socket = NULL; /* Just in case... */
}

static void
init(void)
{
	int	chrooted = 0;

	if (chdir(O.cwd)) {
		syslog(LOG_ERR, "chdir-ing to %s: %m", O.cwd);
		exit(errno == EPERM ? EX_NOPERM : EX_OSFILE);
	}
	syslog(LOG_DEBUG, "chdir-ed to %s", O.cwd);
	if (geteuid() == 0) {
		if (chroot(O.cwd)) {
			syslog(LOG_ERR, "chroot-ing to %s: %m", O.cwd);
			exit(errno == EPERM ? EX_NOPERM : EX_OSFILE);
		}
		syslog(LOG_DEBUG, "chroot-ed to %s", O.cwd);
		chrooted = 1;
	}

	if (O.uid && setuid(O.uid)) {
		syslog(LOG_ERR, "Can't setuid to %ld: %m", (long)O.uid);
		exit(errno == EPERM ? EX_NOPERM : EX_OSERR);
	}
	syslog(LOG_DEBUG, "Continuing as user id %ld", (long)geteuid());
	
	if (O.pidfname) {
		FILE *pidf;

		pidf = fopen(optarg, "w");
		if (pidf == NULL) {
			syslog(LOG_ERR, "Opening the pid "
			    "file %s: %m", optarg);
			exit(errno == EPERM ?
			    EX_NOPERM : EX_CANTCREAT);
		}

		if (fprintf(pidf, "%ld\n", (long)getpid()) < 0) {
			syslog(LOG_ERR, "Writing the pid "
			    "into %s (%p): %m", optarg,
			    pidf);
			exit(EX_IOERR);
		}

		if (fclose(pidf)) {
			syslog(LOG_ERR, "closing %s: %m",
				O.pidfname);
			exit(EX_IOERR);
		}
	}

	if (O.socket) {
		struct stat	sb;
		unsigned int	seconds;

		if (chrooted && strstr(O.socket, O.cwd) == O.socket) {
			O.socket += strlen(O.cwd);
			syslog(LOG_INFO, "Changed the socket path to %s",
			    O.socket);
		}

		for (seconds = 1; lstat(O.socket, &sb) == 0; seconds *= 2) {
			if (sb.st_mode|S_IFSOCK) {
				if (seconds > 16) {
					syslog(LOG_ERR, "Can't wait any longer for %s",
					    O.socket);
					exit(EX_OSFILE);
				}
				syslog(LOG_WARNING, "Socket %s exists"
				    ", waiting %u seconds", O.socket, seconds);
				if (sleep(seconds)) {
					syslog(LOG_ERR, "sleep interrupted (by signal?)"
					    " exiting.");
					exit(EX_TEMPFAIL);
				}
			} else
				syslog(LOG_ERR, "%s exists and is not a "
				    "socket", O.socket);
		}
	}

	if ((O.pidfname || O.socket) && atexit(exiting)) {
		syslog(LOG_ERR, "atexit(): %m");
		exit(EX_SOFTWARE);
	}

}

int
main(int argc, char *argv[])
{
	char	 c;

#ifdef LOG_PERROR
#	define LOG_FLAGS	LOG_PERROR|LOG_PID|LOG_CONS|LOG_NDELAY
#else
/*	Solaris?	*/
#	define LOG_FLAGS	LOG_PID|LOG_CONS|LOG_NDELAY
#endif
	tzset();
	openlog(smfilter.xxfi_name, LOG_FLAGS, LOG_MAIL);
	setlogmask(LOG_UPTO(LOG_INFO)); /* ignore the LOG_DEBUG */

	while ((c = getopt(argc, argv, "Ll:C:g:p:u:bdvhw24r:s:S:B:")) != -1) {
		switch (c) {
		case 'b':
		case 'w':
			init();
			exit(list(argc - optind, argv + optind,
			    c == 'b' ? S_ISGID : S_ISUID));
		case 'd':
			setlogmask(LOG_UPTO(LOG_DEBUG));
			break;
#ifdef SKEM_CLEANUP
		case 'B':
			O.permcleanup = atol(optarg);
			if (O.permcleanup == 0) {
				syslog(LOG_ERR, "the ``-B'' switch needs a"
				    " positive integer number of seconds,"
				    " not ``%s''", optarg);
				exit(EX_USAGE);
			}
			break;
		case 'L':
			/*
			 * If `-l' did not precede `-L', periodic_cleanup()
			 * will return after one shot
			 * XXX periodic_cleanup() must return (void *) to
			 * XXX match the pthread_create() declaration, but
			 * XXX in fact returns an integer exit-code. This
			 * XXX may cause warnings on platforms, where int
			 * XXX is narrower, than void *
			 */
			init();
			/*
			 * skem_cleanup() returns an integer exit-code,
			 * but has to be declared as void* to be accepted
			 * pthread_create()
			 */
			exit((intptr_t)skem_cleanup((void *)1));
		case 'l':
			O.cleanupperiod.tv_sec = atol(optarg);
			if (O.cleanupperiod.tv_sec <= 0) {
				syslog(LOG_ERR, "Invalid cleanup period %s "
				    "-- expecting a positive integer", optarg);
				exit(EX_USAGE);
			}
			break;
#else
		case 'B':case 'L':case 'l':
			syslog(LOG_ERR, "`-%c': cleanup functionality not "
			    "compiled in", c);
			exit(EX_USAGE);
#endif
		case '2':
			smfilter.xxfi_eoh = eoh;
			break;
		case '4':
			O.tempfail = SMFIS_TEMPFAIL;
			/*
			 * Do everything in conn(ect) stage
			 */
			smfilter.xxfi_helo = NULL;
			break;
#ifdef SKEM_LOGWATCHER
		case 'r':
			if (optarg[0] == '-' && optarg[1] == '\0')
				O.reject = NULL;
			else
				O.reject = optarg;
			break;
		case 'S':
			O.spamword = optarg;
			break;
		case 's':
			if (optarg[0] != '-' || optarg[1] != '\0')
				O.ippattern = optarg;
			break;
#else
		case 'r':case 'S':case 's':
			syslog(LOG_ERR, "`-%c': logwatcher functionality not "
			    "compiled in", c);
			exit(EX_USAGE);
#endif
		case 'u':
			/* See, if the numeric uid was given */
			O.uid = atol(optarg);
			if (O.uid == 0 /* root should be rejected anyway */) {
				const struct passwd *pwd;

				pwd = getpwnam(optarg);
				if (pwd == NULL) {
					syslog(LOG_ERR, "Can't parse user %s:"
					    " %m", optarg);
					exit(EX_NOUSER);
				}
				O.uid = pwd->pw_uid;
			}
			break;
		case 'p':
			O.pidfname = optarg;
			break;

		case 'C':
			O.cwd = optarg;
			break;
		case 'g':
			O.grace = atoi(optarg);
			if (O.grace < 3) {
				syslog(LOG_ERR, "requested grace period %s"
				    " (%u) is less than 3 seconds",
				    optarg, O.grace);
				exit(EX_USAGE);
			}
			if (O.grace > 3600*24*5) {
				syslog(LOG_ERR, "requested grace period %s"
				    " (%u seconds) is greater than 5 days",
				    optarg, O.grace);
				exit(EX_USAGE);
			}
			break;
		case 'v':
			fprintf(stderr, "%s\n%s\n%s\n", version,
#ifdef SKEM_LOGWATCHER
			    logwatch_version
#else
			    "no logwatcher compiled in"
#endif
			    ,
#ifdef SKEM_CLEANUP
			    cleanup_version
#else
			    "no cleanup compiled in"
#endif
			);
			exit(0);
		default:
			usage(argv[0], c == 'h' ? 0 : EX_USAGE); /* NORETURN */
		}
	}

	if (optind == argc) {
		syslog(LOG_ERR, "No socketpath specified. Try -h");
		exit(EX_USAGE);
	} else
		O.socket = argv[argc - 1];

	init();

	if (geteuid() == 0) {
		syslog(LOG_ERR, "need not run as root");
		exit(EX_USAGE);
	}

#ifdef SKEM_CLEANUP
	if (O.cleanupperiod.tv_sec &&
	    pthread_create(&O.clean_thread, NULL, skem_cleanup, NULL)) {
		syslog(LOG_ERR, "Could not create the cleanup thread"
		    " requested by the `-c' flag: %m");
		exit(EX_UNAVAILABLE);
	}
#endif
#ifdef SKEM_LOGWATCHER
	if (O.ippattern &&
	    pthread_create(&O.watch_thread, NULL, skem_logwatch, NULL)) {
		syslog(LOG_ERR, "Could not create the logwatch thread"
		    " requested by the `-s' flag: %m");
		exit(EX_UNAVAILABLE);
	}

#else
	/* Nothing to cleanup in certain conditions: */
	if (smfilter.xxfi_eoh == NULL && O.tempfail == SMFIS_TEMPFAIL) {
		/* No cleanup to do in this case */
		smfilter.xxfi_abort = NULL;
		smfilter.xxfi_close = NULL;
	}
#endif

	if (smfi_setconn(O.socket) == MI_FAILURE) {
		syslog(LOG_ERR, "smfi_setconn failed for %s",
		    argv[argc - 1]);
		exit(EX_SOFTWARE);
	}

	if (smfi_register(smfilter) == MI_FAILURE) {
		syslog(LOG_ERR, "smfi_register failed");
		exit(EX_UNAVAILABLE);
	}

	syslog(LOG_INFO, "%s -- starting", version);

	if (O.tempfail == SMFIS_TEMPFAIL) {
		syslog(LOG_INFO, "Will try to reject mail right at the "
		    "connect stage, but it may not work if sendmail was "
		    "not built with the -D_FFR_MILTER_421");
	} else {
		syslog(LOG_INFO, "Will try to reject mail at the helo-stage, "
		    "which is a bit more resource consuming. If sendmail "
		    "was built with -D_FFR_MILTER_421, use the ``-4'' option");
	}

	return smfi_main();
}
